//---------------------------------------------------------------------------

#ifndef PLAH
#define PLAH
                       
typedef list<Term> TermList;

class PLA {
  private:
    unsigned int inputs;        // # of PLA inputs
    unsigned int outputs;       // # of PLA outputs
    PLA_type type;              // PLA type
    bool indexed;        // are indexes valid?

	bool uniq;            // unique terms are present only. TT is generated
	bool keep_TT;			// keeps the TT, even for deleted terms
    TTree TT;               // TTree for uniqueness check.

    TermList data;          // Here are all the terms
    vector<TermList> F;          // on-set of each output
    vector<TermList> D;          // DC-set of each output
    vector<TermList> R;          // off-set of each output

	vector<TermList::iterator> indexterm;		// vector of terms, if PLA is indexed

    vector<string> inames;          // names of input variables
    vector<string> onames;          // names of output variables

    void Init(void);                    // initializes all the values
    void GenerateOnSet( const unsigned int output);    // Generates on-set of the given output. The result is stored in F[output]
    void GenerateOffSet( const unsigned int output);    // Generates off-set of the given output. The result is stored in F[output]
    void GenerateDCSet( const unsigned int output);    // Generates DC-set of the given output. The result is stored in F[output]

	void Reindex(void);                  // reindex the terms, if indexes are not valid

  public:

	PLA(void);
    PLA( const PLA *_pla );			// This is not a copying constructor! Just creates PLA with the same dimensions and type
    PLA( const unsigned int inp, const unsigned int out, const PLA_type _type);
    PLA( const list<Term>& _data, const PLA_type _type );
    ~PLA(void);
    void Clear(void);                   // deallocates all the PLA structures

	// setters & getters
    unsigned int GetInputs(void) const {return inputs;};
    unsigned int GetOutputs(void) const {return outputs;};
    PLA_type GetType(void) const {return type;};
  //  void SetInputs(unsigned int v) {inputs = v;};     // are they needed?
  //  void SetOutputs(unsigned int v) {outputs = v;};     // are they needed?
    void SetType(const PLA_type v);                         // are they needed?
    void SetType(const string v);                         // are they needed?
	string GetIName(const unsigned int i) const;
	string GetOName(const unsigned int i) const;
	void SetIName(const unsigned int i, const string name);
	void SetOName(const unsigned int i, const string name);
    void SetUnique(void);    
    void SetNonUnique(void);
	void KeepTT(void) { keep_TT = true; };			// Sets keeping the TT. Terms from TT will not be deleted when deleting terms. This prevents repeated insertion of a term that is not wanted

    list<Term> *GetData(void);                      // returns pointer to data
	list<Term>::iterator GetTerm(const unsigned int i);	// returns iterator to a term with a given index
    list<Term> *GetOnSet(const unsigned int output);  // returns pointer to on-set. Are they needed?
    list<Term> *GetOffSet(const unsigned int output);  // returns pointer to off-set. Are they needed?
    list<Term> *GetDCSet(const unsigned int output);  // returns pointer to dc-set. Are they needed?

	unsigned int GetIndex( const list<Term>::iterator ti );	// returns index of a term given by iterator

        // data manipulation
    list<Term>::iterator AppendTerm( const Term &t );            // Appends a term to a PLA.
    list<Term>::iterator DeleteTerm(const list<Term>::iterator ti);   // Deletes term pointed by iterator. Returns iterator of the next one
	void AppendPLA( const PLA &_pla );	// appends a given PLA. Terms are copied to the end
	void AppendTermList( const TermList &terms );
	void Purge(void);							// delete unwanted terms
	void MergeTerms(void);				// merges all terms having equal IM
	void MergeNeighbors(void);			// Repeatitively joins neighboring terms (terms differing in one variable only and having equal OM)
	list<Term>::iterator FindTerm( const string t );	// finds a term with equal IM. If not found, returns data.end();

    void LoadPLA(const char *fn);             // Loads PLA from file
    void SavePLA(const char *fn);            // Saves PLA to file
    void SaveVHDL(const char *fn);            // Saves VHDL to file

    void NameVars(void);            // if inames and/or onames are NULL, default names are assigned

        // info functions
    unsigned int GetTerms(void) const { return data.size(); };
    unsigned int GetLiterals(void);
    unsigned int GetOC(void) const;
	unsigned int GetGEs(void);		// Returns GEs*2

      // Special functions
    bool IsImplicantNoOMCheck( const Term& t, const unsigned int output );	// Checks if t is an implicant of the PLA for output. Its value in OM is ignored
	bool IsImplicant( const Term& t, const unsigned int output );				// Checks if t is an implicant of the PLA for output. Its value in OM must be '1'
	bool IsImplicant( const Term& t );				// Checks if t is an implicant of the PLA for all outputs set to '1'.

	void ToF(void);			// Converts PLA to F-type

		// for debug purposes
	void DumpOrigin( const char *fn );

};






//---------------------------------------------------------------------------
#endif
